// Copyright 2020 Northern.tech AS
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

package utils

import "encoding/json"

func AppendStructToMap(Struct interface{}, Map map[string]interface{}) error {
	buf, err := json.Marshal(Struct)
	if err != nil {
		return err
	}
	if err := json.Unmarshal(buf, &Map); err != nil {
		return err
	}

	return nil
}

func MarshallStructToMap(Struct interface{}) (map[string]interface{}, error) {
	retMap := make(map[string]interface{})
	if err := AppendStructToMap(Struct, retMap); err != nil {
		return nil, err
	}
	return retMap, nil
}
